# Contribution Guide

[[_TOC_]]

## General Remarks

A common guideline for source code makes it much easier to understand code
written by other colleagues. In general we adopt the code style guidelines
defined by the OpenFOAM Foundation

* <https://openfoam.org/dev/how-to-contribute>
* <https://openfoam.org/dev/coding-style-guide>
* <https://openfoam.org/dev/code-quality>

## Git Configuration

The Git version control system typically stores its configuration in
`~/.gitconfig`. It is recommended to copy the git configuration into the Docker
containers. When the home directory is mounted to the container, which is the
default when using `foamDockerRun.sh`, nothing needs to be done. Otherwise, use
the following commands to copy the git configuration file into the container

```shell
CONTAINERID=$(docker create -it <image>)
docker cp $HOME/.gitconfig $CONTAINERID:/root/.gitconfig
docker start -i $CONTAINERID
```

For `docker compose`, assuming the service was already started, use the
following command

```shell
docker compose cp $HOME/.gitconfig <service>:/root/.gitconfig
```

**Note:** When using [Visual Studio Code](#integrated-development-environment-visual-studio-code),
the local `~/.gitconfig` is automatically copied into the container.

## Static Code Analysis (Linting)

Make sure that source code and dictionary files contain a valid header:

* for C++ source code: `etc/codeTemplates/app/app.C`
* for dictionaries: `etc/controlDict`
* for Bash: `etc/codeTemplates/script/bash.sh`
* for Python: `etc/codeTemplates/script/python.py`
* for Snakemake: `etc/codeTemplates/script/Snakemake`

If files are an exact or only slightly modified copy of original release files
use the default OpenFOAM Foundation header and mark the changes with code
repository comments (example comments can be found in
use the default OpenFOAM Foundation header and mark the changes with comments
following `etc/codeTemplates/addonCommentStyles`.

### Linting Code Style

Use the `interactive-hook` in `bin/tools` to check for tabs, trailing
whitespaces, and date in copyright headers

```shell
$WM_PROJECT_SITE/$WM_PROJECT_VERSION/bin/tools/interactive-hook -s HEAD -c file1 file2 ...
```

Furthermore, a `pre-commit-hook` is available in `bin/tools`

```shell
ln -s $WM_PROJECT_SITE/$WM_PROJECT_VERSION/bin/tools/pre-commit-hook $WM_PROJECT_SITE/$WM_PROJECT_VERSION/.git/hooks/pre-commit
```

If you are implementing a new model from the literature, make sure to cite it
appropriately in the header file by using the APA style, following the
OpenFOAM Foundation code style guidelines[^1].

### Linting Python, Bash, Markdown, and other

MegaLinter[^2] is used for linting various formats like Python, Shell, JSON,
Markdown and other. To run MegaLinter locally use

```shell
docker run --rm --volume "$PWD:/tmp/lint" registry.hzdr.de/fwdc/infrastructure/megalinter:latest
```

Spellchecking is done with `cspell`[^3] for all strings and comments (included
in MegaLinter). If variables are used inside strings or comments or in the
documentation header they have to be added to the `cspell` dictionary by editing
`etc/cspell-dictionary.txt`. The dictionary can be sorted alphabetically by

```shell
sort -u cspell-dictionary.txt -o cspell-dictionary.txt
```

The unique option `-u` removes duplicates automatically.

## Integrated development environment (Visual Studio Code)

Visual Studio Code[^4] is recommended for both development and general
usage. It comes with an integrated terminal and allows syntax highlighting.

### Setup of Visual Studio Code

1. Install VS Code[^6].
2. To avoid warnings when working on large file systems with VS Code, allow
your local system to watch more files than usual: Open `/etc/sysctl.conf` with
root privileges and add the line `fs.inotify.max_user_watches = 524288`. Apply
these changes with `sudo sysctl -p`.

### Usage of Visual Studio Code

Visual Studio Code (VS Code) settings are provided in a folder named `.vscode/`,
which has to be placed in the working directory.

1. Copy the folder containing the suggested settings to the directory you want
   to open, for example:

   ```bash
   cp -r etc/vscode/ .vscode/
   ```

2. Open the target folder with VS Code. VS Code detects the `.vscode/` folder
   and the settings are loaded.

   If not already done, VS Code asks you to install the recommended extensions
   (defined in `.vscode/extensions.json`).

   **Docker users**:
   * Start your container as usual. Hit `F1` and `Attach to running container`.
     Open the folder `$WM_PROJECT_SITE/$WM_PROJECT_VERSION` with VS Code.
     VS Code detects the `.vscode` folder which was copied in step 1.

### Useful shortcuts for Visual Studio Code

Make sure to provide the VS Code setting files first (described below).

| shortcut       | action                                            |
|----------------|---------------------------------------------------|
| `Ctrl+Shift+B` | build source code (executes `./Allwmake -j`)      |
| `Ctrl+Click`   | go to file/definition                             |
| `Ctrl+P`       | search for specific files to edit them            |
| `Alt+O`        | switch between header and source file             |
| `F1`           | access actions like `Attach to running container` |

### GitLab Workflow with Visual Studio Code

The extension `GitLab Workflow`[^5] provides a large number of GitLab features.
It is already added to the settings suggested above. You have to create and
paste a personal access token (PAT) with `api` and `read_user` permissions.
Afterwards add the following environment variables to your shell:

```bash
export GITLAB_WORKFLOW_INSTANCE_URL="https://codebase.helmholtz.cloud"
export GITLAB_WORKFLOW_TOKEN="PASTE YOUR TOKEN HERE"
```

**Docker users**: Make sure that these variables are also known within the
container.

### Notes on some specific settings for Visual Studio Code

* To make problem matching work, absolute paths are used for compilation:
`Ctoo_VSCode` in `tasks.json` overwrites `Ctoo` defined in `/wmake/rules/*/c++`
* Red zigzag underlines are buggy and therefore disabled:
`"C_Cpp.errorSquiggles": "Disabled"` in `settings.json`

## Unit and Integration Tests

Supply tests for important features in `testAddon` or `testCuda`. A test should:

* run in a few seconds on one core (2 cores for parallel features)
* include an `Allrun` script (template in `etc/template/case`)
* be restricted to utilities and solvers included in software released by the
  OpenFOAM Foundation
* not contain pre- or postprocessing
* contain a `validation` sub-directory for reference data
* contain a `README.md` for documentation
* use `foamPostProcess -func "deviation(field=U, reference=U.ref)"` to check
  against a known numerical or analytical solution

In case the solver or the application writes tabs in output files required for
validation use `sed -i 's/\t/ /g' <file>` to replace tabs by a space and
`diff -b` to ignore differences in the amount of white space. The command
`column -t <file>` can be used to format pretty tables.

Tests in `test` directory are copies of tests in the release and should be
identical.

## Demonstration Tutorials

Supply tutorials for every major development or complex set-up in
`tutorialsAddon` or `tutorialsCuda`. A tutorial should:

* guide users through the features and possibilities of your developments
* run on a maximum of two cores (preferably one)
* include an `Allrun` script (see `etc/template/case` for an example)
* contain a sub-directory called `validation` for reference and additional data

For pre- and postprocessing in tutorials:

* use gnuplot and/or python3
* generate graphs by `( cd validation && ./createGraphs )`
* a template for createGraphs can be found in `etc/codeTemplates/script`
* if available use CSV format with space separator (allows to select column by
  name in gnuplot: `u 1:(column("alpha.air"))`)

Tutorials in `tutorials` directory are copies of tutorials in the release and
should be identical.

---
[^1]: <https://openfoam.org/dev/coding-style-guide/#sec-2-7>
[^2]: <https://oxsecurity.github.io/megalinter/latest>
[^3]: <https://cspell.org>
[^4]: <https://code.visualstudio.com/docs>
[^5]: <https://marketplace.visualstudio.com/items?itemName=GitLab.gitlab-workflow>
[^6]: <https://code.visualstudio.com/docs/setup/setup-overview>
